# -*- coding: utf-8 -*-
#
# Copyright (c) 2019~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from .hash_dict import HashDict
from .hash_set import HashSet

from .comparer import (
    IEqualityComparer,
    ObjectComparer, AnyComparer,
    StringComparers,
)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2019~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from .comparer import IEqualityComparer

def cache(name: str):
    def decorator(func):
        def wrapper(self):
            try:
                return getattr(self, name)
            except AttributeError:
                setattr(self, name, func(self))
            return getattr(self, name)
        return wrapper
    return decorator

class BoxedKey:
    '''
    wrap a key with given equality comparer.
    '''
    __slots__ = ('_comparer', '_obj', '_hashcode', '_cmpvalue')

    def __init__(self, comparer: IEqualityComparer, obj):
        self._comparer = comparer
        self._obj = obj

    @cache('_hashcode')
    def __hash__(self):
        return self._comparer.get_hash(self.get_cmpvalue())

    def __eq__(self, other):
        return self._comparer.equals(self.get_cmpvalue(), other.get_cmpvalue())

    def unbox(self):
        'get the origin object'
        return self._obj

    @cache('_cmpvalue')
    def get_cmpvalue(self):
        return self._comparer.get_cmpvalue(self._obj)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from typing import Set
from collections.abc import MutableSet

from .comparer import IEqualityComparer, ObjectComparer
from ._key import BoxedKey

class HashSet(MutableSet):
    def __init__(self, comparer: IEqualityComparer = None):
        if comparer is None:
            self._comparer = ObjectComparer.instance
        elif not isinstance(comparer, IEqualityComparer):
            raise TypeError('comparer should be IEqualityComparer')
        else:
            self._comparer = comparer

        self._data: Set[BoxedKey] = set()

    def add(self, value):
        return self._data.add(BoxedKey(self._comparer, value))

    def discard(self, value):
        return self._data.discard(BoxedKey(self._comparer, value))

    def __contains__(self, value):
        return BoxedKey(self._comparer, value) in self._data

    def __iter__(self):
        for boxed_key in self._data:
            yield boxed_key.unbox()

    def __len__(self):
        return len(self._data)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
# a comparer like csharp
# ----------

from abc import abstractmethod, ABC

class IEqualityComparer(ABC):
    'the base equality comparer interface'

    def get_cmpvalue(self, obj):
        '''
        get the cmpvalue use for hash and equal compare.
        '''
        return obj

    @abstractmethod
    def get_hash(self, obj):
        '''
        get the hash from object.

        obj is return value of `get_cmpvalue()`
        '''
        raise NotImplementedError

    @abstractmethod
    def equals(self, left, right):
        '''
        compare two values is equals or not.

        `left` and `right` are return value of `get_cmpvalue()`
        '''
        raise NotImplementedError


class ObjectComparer(IEqualityComparer):
    ' default comparer implemention for object '

    instance: IEqualityComparer = None

    def get_hash(self, obj):
        return hash(obj)

    def equals(self, obj1, obj2):
        return obj1 == obj2

ObjectComparer.instance =  ObjectComparer()

class AnyComparer(IEqualityComparer):
    '''
    default comparer implemention for any object,
    use this to compare objects with `is`
    '''

    instance: IEqualityComparer = None

    def get_hash(self, obj):
        return object.__hash__(obj)

    def equals(self, left, right):
        return left is right

AnyComparer.instance = AnyComparer()


class IgnoreCaseStringComparer(IEqualityComparer):
    ' default ignore case comparer implemention for string '

    def get_hash(self, obj: str):
        return hash(obj)

    def equals(self, left, right):
        return left == right

    def get_cmpvalue(self, obj: str):
        return obj.upper()


class StringComparers:
    ' a collection  of comparers for string '

    IgnoreCaseComparer = IgnoreCaseStringComparer()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from typing import Dict
from collections import defaultdict
from collections.abc import MutableMapping

from .comparer import IEqualityComparer, ObjectComparer
from ._key import BoxedKey

class HashDict(MutableMapping):
    '''
    a dict use give comparer.
    if comparer is `None`, use default `ObjectComparer()`
    '''

    def __init__(self, comparer: IEqualityComparer = None):
        if comparer is None:
            self._comparer = ObjectComparer.instance
        elif not isinstance(comparer, IEqualityComparer):
            raise TypeError('comparer should be IEqualityComparer')
        else:
            self._comparer = comparer

        self._data: Dict[BoxedKey, object] = {}

    def __getitem__(self, key):
        return self._data[BoxedKey(self._comparer, key)]

    def __setitem__(self, key, value):
        self._data[BoxedKey(self._comparer, key)] = value

    def __delitem__(self, key):
        del self._data[BoxedKey(self._comparer, key)]

    def __iter__(self):
        for boxed_key in self._data:
            yield boxed_key.unbox()

    def __len__(self):
        return len(self._data)

    def setdefault(self, key, value):
        '''
        Insert key with a value of default if key is not in the dictionary.

        Return the value for key if key is in the dictionary, else default.
        '''
        # overwrite to ensure this is atomic
        return self._data.setdefault(BoxedKey(self._comparer, key), value)

    def copy(self):
        'Make a shallow copy'
        new_one = HashDict(self._comparer)
        new_one.update(self)
        return new_one

    @staticmethod
    def fromkeys(iterable, value=None, *, comparer: IEqualityComparer = None):
        'Create a new dictionary with keys from iterable and values set to value.'
        new_one = HashDict(comparer)
        for item in iterable:
            new_one[item] = value
        return new_one
